import inspect
from typing import *
from starlette.routing import Route as _Route, Request, PlainTextResponse
from starlette.responses import Response
from ..tools import CSTYLE, get_calling_app
from .. import logger, exceptions


__all__ = [
    'Route',
    'registered',
    'route',
    'append',
    'abs_url',
    'format_path',
    'static_routes_prefixes',
    'is_static_path'
]


class Route(_Route):
    """ The wrapper class extending the Starlette's Route class """

    def __init__(
            self,
            path: str,
            endpoint: Callable,
            *,
            methods: List[str] = None,
            name: str = None,
            **kwargs
    ):
        super().__init__(
            path,
            self._decorate_endpoint(endpoint),
            methods=methods,
            name=name,
            **kwargs
        )

    def _decorate_endpoint(self, endpoint: Callable) -> Callable:
        """ Decorating the endpoint (if the endpoint is a function or method) with
        extra functionality. """

        if inspect.isclass(endpoint):
            # If the endpoint is the HTTPEndpoint class - return is
            # straight without any decoration
            return endpoint

        # Decorate the endpoint function with some extra functionality
        async def _decorated(request: Request) -> Response:
            try:
                return await endpoint(request)

            except exceptions.AccessDenied:
                return PlainTextResponse("Forbidden", status_code=403)

            except exceptions.NotAuthenticated:
                return PlainTextResponse("Unauthorized", status_code=401)

            except exceptions.ObjectNotFoundError as e:
                return PlainTextResponse(e.details, status_code=404)

            except exceptions.ApiError as e:
                return PlainTextResponse(e.details, status_code=e.status_code)

            except exceptions.DatabaseError as e:
                return PlainTextResponse(e.details, status_code=e.status_code)

        return _decorated


registered: List[Route] = []
static_routes_prefixes: List[str] = [
    '/static'
]


def append(r: Route) -> None:
    """ Appends the given route to the routing table of the project."""

    _methods: str = ','.join(str(s) for s in r.methods)
    logger.debug(f"routed {CSTYLE['white']}{_methods}{CSTYLE['clear']} {CSTYLE['pink']}{r.path}{CSTYLE['clear']}")
    registered.append(r)


def abs_url(path: str, app: Optional[str] = None) -> str:
    """ Returns the absolete URL for the given path and app. If the app
    has been ommited - use the calling app by default. """

    app = app or get_calling_app()
    if path.startswith('//'):
        return path[1:]
    if path == '/':
        path = ''
    return '/' + ('/'.join([s for s in (app, path.lstrip('/')) if s])).lstrip('/')


def format_path(
        relpath: str,
        prefix: Optional[str] = None,
        inner: Optional[str] = None,
        suffix: Optional[str] = None
) -> str:
    if relpath.startswith('//'):
        return relpath[1:]
    route_app: str = get_calling_app()
    prefix: str = '' if not prefix else prefix.strip('/')
    inner: str = '' if not inner else inner.strip('/')
    suffix: str = '' if not suffix else suffix.strip('/')
    return f"/{'/'.join([s for s in [prefix, route_app, inner, relpath.strip('/'), suffix] if s])}".\
        rstrip('/').replace('//', '/')


def route(path: str, methods: Optional[List[str]] = None) -> Callable:
    """ Decorator used to route the function endpoint. It appends the resulting
    route to the routing table of the project. """

    def decorator(endpoint: Callable) -> Callable:
        request_methods = methods or ['GET']
        append(Route(format_path(path), endpoint, methods=request_methods))
        return endpoint
    return decorator


def is_static_path(p: str) -> bool:
    """ Returns `True` if the given FQ path is the static one (endpoints to
    the static file storage, not the response function or class), and `False`
    otherwise. """

    p: str = p.lstrip('/')
    if '/' in p:
        return False
    p: list = p.split('.')
    if len(p) != 2:
        return False
    if len(p[0]) == 0:
        return False
    if len(p[1]) == 0:
        return False
    return True






class HTTPResponseError(BaseException):
    pass


class NotFoundResponse(HTTPResponseError):
    pass


class AAAError(BaseException):
    pass


class NotAuthenticated(AAAError):
    pass


class AccessDenied(AAAError):
    pass


class AuthenticationFailed(AAAError):
    pass


class ApiError(BaseException):
    def __init__(self, http_code: int = 400, details: str = "API error"):
        self.status_code: int = http_code
        self.details: str = details


class ObjectNotFoundError(BaseException):
    def __init__(self, details: str = None):
        from .l10n import gettext
        self.details: str = details or gettext("The specified object was not found", 'system.messages')


class DatabaseError(BaseException):
    def __init__(self, http_code: int = 400, details: str = "Database error"):
        self.status_code: int = http_code
        self.details: str = details


class DatabaseIntegrityError(DatabaseError):
    def __init__(self, details: str = None) -> None:
        from .l10n import gettext
        super().__init__(
            409,
            details or gettext(
                "Changes cannot be made because they conflict with other existing properties or objects.",
                'system.messages'
            )
        )




import types
import asyncio
from typing import *
from wefram import tools, runtime


def print_help() -> None:
    print("")


async def test_run() -> None:
    from .routines.project import ensure_apps_loaded

    ensure_apps_loaded()


async def test_app(command: str, args: List[str]) -> None:
    from wefram import apps

    app: str = command
    if app != 'system' and app != 'wefram' and not apps.is_enabled(app):
        print(f"app '{app}' is not enabled and cannot be tested")
        return

    tests: types.ModuleType = tools.load_app_module(app, 'tests')
    if tests is None:
        print(f"app '{app}' has no module named 'tests' (tests.py) to be ran within test")
        return

    test_name: str = args.pop(0) if len(args) else 'main'
    if not hasattr(tests, test_name):
        print(f"app '{app}' has no test named '{test_name}' in the 'tests' module")
        return

    test = getattr(tests, test_name)
    if asyncio.iscoroutinefunction(test):
        await test(*args)
    else:
        test(args)


async def run(args: List[str]) -> None:
    if not args:
        print_help()
        return

    command = args.pop(0)

    if command == 'run':
        print("(test run)")
        print("vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv")
        print("")

        await test_run()

        print("")
        print("^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^")
        return

    else:
        await runtime.within_cli(test_app, command, args)
        return
